/*
 * realreplay lib
 *
 * see http://sed.free.fr/realplay for the documentation.
 * This is public domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <bfd.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <asm/fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <stdarg.h>

int (*the_open)(const char *, int, mode_t) = 0;
int (*the_ioctl)(int, int, ...);
int (*the_close)(int) = 0;
//int  (*the_select)(int, fd_set *, fd_set *, fd_set *, struct timeval *) = 0;

/* some local declarations */
static void emergency_exit(int);

/* our static globals */
struct trace {
  void *address;
  size_t address_size;
  int size;
  int max_size;
  void **bt;
};

static unsigned int lib_init=0;

static void init_lib(void)
{
  char *libc_path;
  void *p;

  /* useless check, but good practice */
  if (lib_init)
    return;

  lib_init = 1;

  libc_path = getenv("LIBC_PATH");
  if (!libc_path) {
    fprintf(stderr, "error : LIBC_PATH environment variable not declared\n");
    emergency_exit(1);
  }

  /* let's interface with the libc */
  p = dlopen(libc_path, RTLD_LAZY);
  if (!p) {
    fprintf(stderr, "dlopen error. You surely gave a wrong LIBC_PATH\n");
    emergency_exit(1);
  }

  the_open = dlsym(p, "open");
  if (!the_open) {
    perror("dlsym(open) error");
    exit(1);
  }

  the_close = dlsym(p, "close");
  if (!the_close) {
    perror("dlsym(close) error");
    exit(1);
  }

  the_ioctl = dlsym(p, "ioctl");
  if (!the_ioctl) {
    perror("dlsym(ioctl) error");
    exit(1);
  }
}

int snd=-1;

int close(int fd)
{
  if (!the_close)
    init_lib();
  if (!the_close) {
    fprintf(stderr, "problem with preload stuff\n");
    return -1;
  }
  if (fd == snd) {
    snd = -1;
    fprintf(stderr, "closing soundcard\n");
  }
  return the_close(fd);
}

int open(const char *t, int z, mode_t r)
{
  fprintf(stderr, "open %s\n", t);
  if (!the_open)
    init_lib();
  if (!the_open) {
    fprintf(stderr, "problem with preload stuff\n");
    return -1;
  }
  if (!strcmp(t, "/dev/mixer")) {
    int fd = the_open("/tmp/sed.pipo.mixer", O_RDWR|O_CREAT|O_TRUNC, 0777);
    fprintf(stderr, "catching up mixer %d\n", fd);
    return fd;
  }
  if (!strcmp(t, "/dev/dsp")) {
    int fd = the_open("/tmp/sed.pipo", O_RDWR|O_CREAT|O_TRUNC, 0777);
    fprintf(stderr, "/dev/dsp detected -> opening /tmp/sed.pipo %d %d-> %d\n", z, r, fd);
    snd = fd;
    return fd;
  }
  return the_open(t, z, r);
}

typedef struct audio_buf_info {
   int fragments; 
int fragstotal; /* Total # of fragments allocated */
int fragsize;   /* Size of a fragment in bytes */
int bytes;   /* Available space in bytes (includes partially used fragments) */
/* Note! 'bytes' could be more than fragments*fragsize */
} audio_buf_info;

int ioctl(int d, unsigned long request, ...)
{
  va_list ap;
  void *r;
  if (!the_ioctl)
    init_lib();
  if (!the_ioctl) {
    fprintf(stderr, "problem with preload stuff\n");
    return -1;
  }
  va_start(ap, request);
  r = va_arg(ap, void *);
  va_end(ap);
  if (d == snd && snd != -1) {
    char c = ((unsigned)request >>8) & 0xff;
    int w = (unsigned)request & 0xff;

    if (c == 'P' && w == 0) {
      //fprintf(stderr, "reset card\n");
    } else
    if (c == 'P' && w == 2) {
      fprintf(stderr, "set speed at %d\n", *(int *)r);
    } else
    if (c == 'P' && w == 6) {
      fprintf(stderr, "set nb channels %d\n", *(int *)r);
    } else
    if (c == 'P' && w == 5) {
      fprintf(stderr, "set format %d\n", *(int *)r);
    } else
    if (c == 'P' && w == 10) {
      //fprintf(stderr, "set nb fragments %8.8x\n", *(int *)r);
    } else
    if (c == 'P' && w == 12) {
      audio_buf_info *x = r;
      //fprintf(stderr, "asking for space ? ok, there we go :)\n");
      x->fragments = 10;
      x->fragstotal = 10;
      x->fragsize = 4096;
      x->bytes = 10 * 4096;
    } else
    if (c == 'P' && w == 23) {
      //fprintf(stderr, "getodelay %8.8x\n", *(int *)r);
      *(int *)r = 0000;
    } else {
      fprintf(stderr, "unkown ioctl\n");
      fprintf(stderr, "calling ioctl on %d request=%8.8x\n", d, (unsigned)request);
      fprintf(stderr, "%c %d\n", ((unsigned)request >>8) & 0xff, (unsigned)request & 0xff);
      return -1;
    }
    return 0;
  }
  return the_ioctl(d, request, r);
}

static void emergency_exit(int status)
{
  kill(getpid(), SIGTERM);
}
